# 🥋 ploson

#### Programming Language On Serializable Object Notation (JSON)

> Build your own DSL and have contracts as JSON documents saved in a database.

There are many "JSON Lisp" type of packages, but they're often too narrow in functionality and have a lengthy and arduous syntax meant to be generated rather than manually written. The point with these type of modules is often to enable domain specific languages (DSL) with a DB friendly syntax (JSON) - and so is Ploson.

- Asynchronous & concurrent async support
- Human developer friendly (made to easily write by hand)
- Fully customizable scope of available environment identifiers
  - Build your own API
- Security guards for safe use as DSL
  - `eval` & other unsafe functions forbidden
- Simple, yet competent
- Plugin based implementation

## Primitives

```js
ploson(2); // -> 2
ploson(3.14); // -> 3.14
ploson(true); // -> true
ploson(null); // -> null
ploson(undefined); // -> undefined
```

### Strings

```js
ploson('`hello world`'); // -> "hello world"
// Would be considered an identifier without ``.
```

## Arrays: running functions

```js
// Highly recommended to add e.g. ramda, and run like:
ploson(['R.add', 1, 2]); // -> 3
// Easily create arrays with built in 'of' method:
ploson(['of', 1, true, '`foo`']); // -> [1, true, "foo"]
// Create a Date object with time now (requires defaultEnv):
ploson(['newDate']); // -> Mon Jul 05 2021 19:41:35 GMT+0200 (Central European Summer Time)
```

## Objects:

- Parallel async
- Returns the object, evaluated
- Will put key/value pairs in the variable scope as an automatic side effect

```js
ploson({ a: 1 }); // -> { a: 1 }

// See "Built-in functions" below about ','
ploson([
  ',',
  { user: ['fetchProfile', '$uId'], conf: ['fetchConfiguration'] }, // parallell async
  { name: ['R.propOr', '$conf.defaultName', 'name', '$user'] },
]);
// -> { name: 'John Doe' }

// In the above example all 3 keys, with values, was added to the variable scope,
// accessible through $user, $conf, $name.
```

To make the second example above work, it is set up like this:

```js
import { createRunner, defaultEnv, envPlugin, evaluatePlugin, varsPlugin } from 'ploson';
import * as R from 'ramda';
import * as D from 'date-fns/fp';

const ploson = createRunner({
  staticEnv: { ...defaultEnv, R, D, fetchProfile, fetchConfiguration },
  plugins: [envPlugin(), evaluatePlugin(), varsPlugin({ vars: { uId: 'john@doe.ex' } })],
});
```

You only need the `plugins:` row **if** you want to populate the variable scope beforehand.
`uId` could also have been added to `staticEnv`, and referenced without prefix `$`.

## Built-in functions

There are a few built-in functions in the predefined environment scope.

| Function name | Implementation             | Comment                                                     |
| ------------- | -------------------------- | ----------------------------------------------------------- |
| `console`     | `console`                  | We should always be able to log to debug.                   |
| `lastArg`     | `R.pipe(Array.of, R.last)` | Evaluate the arguments and simply return the last one       |
| `,`           |                            | Same as `lastArg` (see JS's comma operator)                 |
| `of`          | `Array.of`                 | We should always be able to create arrays                   |
| `void`        | `() => undefined`          | Could be used like lastArg/, but without returning anything |

_You have to override them if you would like to remove them._

## Security

Blocked identifiers:

- eval
- Function
- constructor
- setTimeout, setInterval
  - Accepts function body as sting just like Function constructor & eval.

## `defaultEnv`

As seen in an example above, we can add `defaultEnv` to `staticEnv` to populate the environment scope with a bunch of basic JavaScript constructs:

| Function name | Comment                             |
| ------------- | ----------------------------------- |
| `undefined`   |
| `Array`       | Access native Array API.            |
| `Object`      | Access native Object API.           |
| `String`      | Access native String API.           |
| `Number`      | Access native Number API.           |
| `Promise`     | Access native Promise API.          |
| `newPromise`  | Create Promises.                    |
| `Date`        | Access native Date API.             |
| `newDate`     | Create Dates.                       |
| `Math`        | Access native Math API.             |
| `parseInt`    | The native `parseInt` function.     |
| `parseFloat`  | The native `parseFloat` function.   |
| `Set`         | Access native Set API.              |
| `Map`         | Access native Map API.              |
| `newSet`      | Create Sets.                        |
| `newMap`      | Create Maps.                        |
| `RegExp`      | Create RegExps & access RegExp API. |
| `fetch`       |

## Plugins

This is the plugin sequence that you get by default, without specifying the `plugins` key:

```js
const ploson = createRunner({
  plugins: [envPlugin(), evaluatePlugin(), varsPlugin()],
});
```

If you want to add or modify on the plugin level, just modify the above `plugins` line.

Plugins can return any or all of these 3:

```js
{
  staticEnv: { /* ... */ },
  onEnter: ({ state, current, originalNode, envHas, getFromEnv }) => {},
  onLeave: ({ state, current, originalNode, node, envHas, getFromEnv }) => {},
}
```

Both `onEnter` and `onLeave` functions should return either `undefined` or an **event**:
`{ type 'REPLACE', current: X }` **OR** `{ type: 'ERROR', error: Error('MSG') }`

> Recommended to use **`onLeave`** over `onEnter`.

# Licence

Hippocratic License Version 2.1
