import * as R from 'ramda';
import { hasPath, mapAsync } from 'rambdax';
// import set from 'lodash.set';
import isPlainObject from 'lodash.isplainobject';
// import deepFreeze from 'deep-freeze-strict';

import { OBJ, getPath, ensureError, parallelAwaitObjectVals } from './utils';
import { envPlugin } from './env';
import { evaluatePlugin } from './evaluate';
import { varsPlugin } from './vars';

const pluginEnvs = R.pipe(R.pluck('staticEnv'), R.mergeAll);

const promiseWrapperKey = '%_PROMISE_OBJECT_%';
const maybeWrap = R.when(R.is(Promise), R.objOf(promiseWrapperKey));
const maybeUnwrap = (node) => {
  if (!isPlainObject(node)) return node;
  const keys = Object.keys(node);
  return keys.length === 1 && keys[0] === promiseWrapperKey ? node[promiseWrapperKey] : node;
};

const UNTOUCHED = 'untouched';
const REPLACED = 'replaced';
const ERROR = 'error';

const pluginCallbackReducer =
  (ctx) =>
  ({ state, current, error }, fn) => {
    if (state === ERROR) return { state, error };
    const iterObj = fn({ ...ctx, state, current });
    if (iterObj?.type === 'ERROR') return { state: ERROR, error: iterObj?.error };
    if (iterObj?.type === 'REPLACE') return { state: REPLACED, current: iterObj?.current };
    return { state, current };
  };

export const createRunner = ({
  staticEnv = OBJ,
  // scopes = OBJ,
  plugins = [envPlugin(), evaluatePlugin(), varsPlugin()],
} = OBJ) => {
  const env = { ...pluginEnvs(plugins), ...staticEnv }; // TODO: deepFreeze?
  const envHas = (pth) => hasPath(pth, env);
  const getFromEnv = getPath(env);
  const onEnterFns = R.pipe(R.pluck('onEnter'), R.filter(R.is(Function)))(plugins);
  const onLeaveFns = R.pipe(R.pluck('onLeave'), R.filter(R.is(Function)))(plugins);

  const processNode = async (originalNode) => {
    // TODO: freeze originalNode
    // ENTER
    const onEnterCtx = R.reduce(pluginCallbackReducer({ originalNode, envHas, getFromEnv }), {
      state: UNTOUCHED,
      current: originalNode,
    })(onEnterFns);

    if (onEnterCtx.state === ERROR) return ensureError(onEnterCtx.error);
    /* eslint-disable fp/no-let, fp/no-mutation */
    let node = onEnterCtx.current;
    // const node = originalNode;

    // RECURSE
    if (Array.isArray(originalNode) || isPlainObject(originalNode)) {
      node = await mapAsync(processNode, node);
      node = R.map(maybeUnwrap)(node);
    }
    if (R.is(Error, node)) return node;

    // PARALLEL AWAIT POSSIBLE PROMISES
    if (isPlainObject(node)) node = await parallelAwaitObjectVals(node);
    /* eslint-enable fp/no-let, fp/no-mutation */
    // if (R.is(Error, halfEvaluated)) return halfEvaluated; // CAN THIS EVER HAPPEN?!

    // console.log('halfEvaluated:', halfEvaluated);

    // LEAVE
    const onLeaveCtx = R.reduce(pluginCallbackReducer({ originalNode, node, envHas, getFromEnv }), {
      state: onEnterCtx.state,
      current: node,
    })(onLeaveFns);

    const unfoundIdentifier = R.is(String, onLeaveCtx.current) && onLeaveCtx.state === UNTOUCHED;

    if (onLeaveCtx.state === ERROR) return ensureError(onLeaveCtx.error);
    if (unfoundIdentifier) return Error(`Unfound identifier: ${onLeaveCtx.current}`);

    return maybeWrap(onLeaveCtx.current);
  };

  return R.pipe(processNode, maybeUnwrap);
};
